import six
from docker_emperor.commands import Command
from docker_emperor.nodes import HasEnvironment, HasServices
from docker_emperor.utils import combine, memoized_property, memoized
import docker_emperor.logger as logger


__all__ = ['HasMachines', 'Machines', 'Machine']


# DRIVERS

# Amazon Web Services
# Microsoft Azure
# Digital Ocean
# Exoscale
# Google Compute Engine
# Generic
# Microsoft Hyper-V
# OpenStack
# Rackspace
# IBM Softlayer
# Oracle VirtualBox
# VMware vCloud Air
# VMware Fusion
# VMware vSphere
# VMware Workstation (unofficial plugin, not supported by Docker)
# Grid 5000 (unofficial plugin, not supported by Docker)


class HasMachines():
    
    @property
    def machines(self):
        attr = '_machines'
        if not hasattr(self, attr): setattr(self, attr, Machines(self.data.pop('machines', {})))
        return getattr(self, attr)


class Machines():

    def __init__(self, data):
        self.data = data
        if not isinstance(self.data, dict): self.data = {}

    @memoized
    def select(self, name='default', default_data=None):
        data = self.data.get(name, default_data)
        if isinstance(data, six.string_types):
            name = data
            data = self.data.get(name)
        return Machine(name, data)

    @property
    def default(self): 
        return self.select('default', default_data={})
    
    @property
    def all(self): 
        return { name: Machine(name, data) for name, data in self.data.items() }


class Machine(HasEnvironment, HasServices):

    COMMANDS = [
        'ssh'
    ]

    def __init__(self, name, data, path="docker-machine"):

        self.key = name
        self.name = name
        self.data = data
        if not isinstance(self.data, dict): self.data = {}
        self.path = path

        # SET DRIVER
        self.driver = self.data.get('driver')
        if not isinstance(self.driver, six.string_types): self.driver = 'generic --generic-ip-address localhost'

        # SET HOSTS
        self.hosts = self.data.get('hosts')
        if not isinstance(self.hosts, list): self.hosts = []

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.name)

    @classmethod
    def has_command(cls, name):
        return hasattr(cls, 'command_{}'.format(name))

    def run_command(self, name, *args):
        if Machine.has_command(name):
            getattr(self, 'command_{}'.format(name))(*args)

    def _run(self, *args, **kwargs):
        kwargs['tty'] = kwargs.get('tty', True)
        return Command(
            self.path, 
            *args, 
            **kwargs
        )

    @property
    def exists(self):
        cmd = self._run("ls", "--filter", "NAME=" + self.name, "--format", "{{.Name}}", env=self.docker_env, tty=False)
        for line in cmd.lines:
            if line == self.name:return True
        return False

    def command_ssh(self, *args, **kwargs):
        self._run('ssh', self.name, *args, **kwargs)

    def command_active(self, *args, **kwargs):
        self._run('active', *args, env=self.docker_env, **kwargs)

    def command_exists(self, *args, **kwargs):
        logger.warning(('Machine {} exists' if self.exists else 'Machine {} doesn\'t exists, execute: dw create').format(self.name))


    @property
    def docker_env(self):
        cmd = Command(
            self.path, 
            'env', 
            self.name, 
            log=True
        )
        starts = 'export '
        return [line.lstrip(starts) for line in cmd.lines if line.startswith(starts)]


    @property
    def start(self):
        Command(self.path, 'start', self.name, sys=True).log()    

    @property
    def status(self):
        return Command(self.path, 'status', self.name).out
    
    @property
    def ip(self):
        return Command(self.path, 'ip', self.name, log=False, env=self.docker_env).out.strip()

    @property
    def pwd(self):
        return Command(self.path, 'ssh', self.name, 'pwd', log=False, env=self.docker_env).out.strip()

    @property
    def inspect(self):
        cmd = self._run('inspect', self.name, env=self.docker_env, tty=False)
        return cmd.out

    @property
    def active(self):
        cmd = self._run('active', env=self.docker_env, tty=False)
        return cmd.out


    def create(self):
        cmd = self._run('create', '--driver', self.driver, self.name, env=self.docker_env)
        return cmd

    def remove(self):
        cmd = self._run('rm', self.name, env=self.docker_env)
        return cmd

# active
# config
# create
# env
# help
# inspect
# ip
# kill
# ls
# mount
# provision
# regenerate-certs
# restart
# rm
# scp
# ssh
# start
# stop
# upgrade
# url


import six
import os
import subprocess
import docker_emperor.logger as logger


__all__ = ['Command']

COMPOSE_COMMANDS = [
    'build',
    'bundle',
    'config',
    'create',
    'down',
    'events',
    'exec',
    'help',
    'kill',
    'logs',
    'pause',
    'port',
    'ps',
    'pull',
    'push',
    'restart',
    'rm',
    'run',
    'scale',
    'start',
    'stop',
    'top',
    'unpause',
    'up',
]

MACHINE_COMMANDS = [
    'active',
    'config',
    'create',
    'env',
    'help',
    'inspect',
    'ip',
    'kill',
    'ls',
    'mount',
    'provision',
    'regenerate-certs',
    'restart',
    'rm',
    'scp',
    'ssh',
    'start',
    'status',
    'stop',
    'upgrade',
    'url',
]



class Command():

    def __repr__(self):
        return self.cmd_line

    def __init__(self, *args, **kwargs):

        self.cmd = list(args)
        self.error = None

        for i, arg in enumerate(self.cmd):
            if isinstance(arg, six.integer_types):
                self.cmd[i] = str(arg)
            elif not isinstance(arg, six.string_types):
                raise Exception("Argument {} is invalid: {} is not string types".format(i, arg))

        # if kwargs.get('log', True):
        #     logger.info(" ".join(self.cmd))

        self.env = kwargs.pop('env', [])
        if not self.env: self.env = []
        if not isinstance(self.env, list):
            raise Exception("Env arguments are invalid: {} is not list types".format(self.env))       

        self.cmd_line = " ".join(self.env + self.cmd)

        # if kwargs.get('log', True):
        #     logger.info(" ".join(self.env + self.cmd))

        # if kwargs.get('log', True):
        #     logger.comment("Env.\n" + "\n".join(self.env))

        if kwargs.get('sys', False):
            self.out = os.system(" ".join(self.env + self.cmd))

        else:
            self.lines = []
            self.process = subprocess.Popen(
                " ".join(self.env + self.cmd), 
                shell=True, 
                stdin=subprocess.PIPE, 
                stdout=subprocess.PIPE, 
                stderr=subprocess.PIPE
            )
            self.out = ''

            for line in iter(self.process.stdout.readline, ''):  # replace '' with b'' for Python 3
                self.out += line
                self.lines.append(line.strip())
                self.last_line = line

            self.out = self.out.strip()
            stdout, stderr = self.process.communicate()
            self.error_code = self.process.returncode
            self.error = stderr.decode('utf-8').strip()


        if self.error:                
            self.error = self.error.replace('docker-machine ls', 'dw ls')
            if not kwargs.get('silently_fail', False):
                logger.error(self.error)
                exit(0) 

    def log(self):
        logger.success(self.out)


    # def _run(self, cmd, raise_error=True):
    #     """
    #     Run a docker-machine command, optionally raise error if error code != 0
    #     Args:
    #         cmd (List[str]): a list of the docker-machine command with the arguments to run
    #         raise_error (bool): raise an exception on non 0 return code
    #     Returns:
    #         tuple: stdout, stderr, error_code
    #     """
    #     cmd = [self.machine_path] + cmd
    #     print(" ".join(cmd))

    #     # output = subprocess.check_output(cmd)
    #     # print(output.strip())
    #     # return output

    #     process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    #     for line in iter(process.stdout.readline, ''):  # replace '' with b'' for Python 3
    #         print(line.strip())

    #     stdout, stderr = process.communicate()
    #     # print(stdout, stderr)
    #     error_code = process.returncode
    #     # if raise_error and error_code:
    #     #     raise RuntimeError("cmd returned error %s: %s" % (error_code, stderr.decode('utf-8').strip()))
    #     return stdout.decode('utf-8'), stderr.decode('utf-8'), error_code


    # def _run_blocking(self, cmd, raise_error=True):
    #     """
    #     Run a docker-machine command, optionally raise error if error code != 0
    #     Args:
    #         cmd (List[str]): a list of the docker-machine command with the arguments to run
    #         raise_error (bool): raise an exception on non 0 return code
    #     Returns:
    #         tuple: stdout, stderr, error_code
    #     """
    #     cmd = [self.machine_path] + cmd

    #     # output = subprocess.check_output(cmd)
    #     # print(output.strip())
    #     # return output


    #     p = subprocess.open(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    #     stdout, stderr = p.communicate()
    #     error_code = p.wait()

    #     # if raise_error and error_code:
    #     #     raise RuntimeError("cmd returned error %s: %s" % (error_code, stderr.decode('utf-8').strip()))
    #     return stdout.decode('utf-8'), stderr.decode('utf-8'), error_code


import os
from docker_emperor.commands import Command
import docker_emperor.logger as logger


def run(root, *args, **kwargs):

    machine_status = root.machine.status
    if machine_status == 'Stopped':
        root.machine.start()

    if machine_status == 'Running':

        root.run_command('down')
        root.run_command('up')


import os
from docker_emperor.nodes import HasEnvironment
from docker_emperor.utils import combine, memoized_property, yaml


__all__ = ['HasServices', 'Services', 'Service']


class HasServices():

    preserve_services = True

    @memoized_property
    def services(self):
        if self.preserve_services:
            self.data['services'] = Services(self.data.get('services', {}))
            return self.data['services']
        else:
            return Services(self.data.pop('services', {}))


class Services(dict):

    def __new__(cls, *args, **kwargs):
        return dict.__new__(cls, *args, **kwargs)

    def __init__(self, data):
        if not isinstance(data, dict): data = {}
        super(Services, self).__init__(data)
        for name, data in self.items(): self[name] = Service(name, data)

    def __gt__(self, services):
        if not isinstance(services, Services): return self
        return services < self

    def __lt__(self, services):
        if not isinstance(services, Services): return self
        for service in services:
            service < self.get(service.name)
        return self

    def __iter__(self):
        for name, service in self.items():
            yield service

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, super(dict, self).__repr__())

    @property
    def data(self): return self

def services_representer(dumper, services):
    return dumper.represent_dict(services)
yaml.add_representer(Services, services_representer)


class Service(HasEnvironment, dict):

    preserve_environment = True

    def __new__(cls, *args, **kwargs):
        return dict.__new__(cls, *args, **kwargs)

    def __init__(self, name, data):
        self.name = name
        if not isinstance(data, dict): data = {}
        super(Service, self).__init__(data)

    def __gt__(self, service):
        if not isinstance(service, Service): return self
        return service < self

    def __lt__(self, service):
        if not isinstance(service, Service): return self
        self.__init__(self.name, combine(service, self))
        return self

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.name)

    @property
    def data(self): return self
    

def service_representer(dumper, service):
    return dumper.represent_dict(service)
yaml.add_representer(Service, service_representer)

#print(Service('test1', { 'key1-test1': 1, 'key2-test1': 1}) < Service('test2', { 'key1-test1': 2, 'key3-test2': 1}))
# print(yaml.dump(Service('test1', { 'key1-test1': 1, 'key2-test1': 1})))


import six
import yaml
try:
    from yaml import CLoader as YamlLoader, CDumper as YamlDumper
except ImportError:
    from yaml import Loader as YamlLoader, Dumper as YamlDumper

__all__ = ['memoized_property', 'memoized', 'yamp_load', 'yamp_dump', 'yaml' ]


def yamp_dump(data):
    return yaml.dump(data, Dumper=YamlDumper, default_flow_style=False, indent=4)

def yamp_load(content):
    return yaml.load(content, Loader=YamlLoader)

        
class memoized_property(object):

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        if doc is None and fget is not None and hasattr(fget, "__doc__"):
            doc = fget.__doc__
        self.__get = fget
        # self.__set = fset
        # self.__del = fdel
        # self.__doc__ = doc
        if fget is not None:
            self._attr_name = '___'+fget.func_name
    
    def __get__(self, inst, type=None):
        if inst is None:
            return self
        if self.__get is None:
            raise AttributeError, "unreadable attribute"
        
        if not hasattr(inst, self._attr_name):
            result = self.__get(inst)
            setattr(inst, self._attr_name, result)
        return getattr(inst, self._attr_name)
    
    # def __set__(self, inst, value):
    #     if self.__set is None:
    #         raise AttributeError, "can't set attribute"
    #     delattr(inst, self._attr_name)
    #     return self.__set(inst, value)

    # def __delete__(self, inst):
    #     if self.__del is None:
    #         raise AttributeError, "can't delete attribute"
    #     delattr(inst, self._attr_name)
    #     return self.__del(inst)

class memoized(object):

    def __init__(self, function):
        self._function = function
        self._cacheName = '_cache__' + function.__name__
    def __get__(self, instance, cls=None):
        self._instance = instance
        return self
    def __call__(self, *args, **kwargs):
        cache = self._instance.__dict__.setdefault(self._cacheName, {})
        if cache.has_key(args):
            return cache[args]
        else:
            object = cache[args] = self._function(self._instance, *args, **kwargs)
            return object
# class Combination():

#     def __gt__(self, elm2):

#         elm1 = self

#         # None - None
#         if elm1 is None and elm2 is None:
#             return None

#         # None - str
#         elif elm1 is None and isinstance(elm2, six.string_types):
#             elm1 = elm2

#         # None - dict
#         elif elm1 is None and isinstance(elm2, dict):
#             elm1 = elm2
#             combine(elm1, elm2)

#         # None - List
#         elif elm1 is None and isinstance(elm2, list):
#             elm1 = elm2
#             combine(elm1, elm2)

#         # Dict - Dict
#         elif isinstance(elm1, dict) and isinstance(elm2, dict):
#             for key2, val2 in elm2.items():
#                 val1 = elm1.get(key2, val1)
#                 if val1:
#                     elm1[key2] = val1

#         # List - List
#         elif isinstance(elm1, list) and isinstance(elm2, list):
#             elm1 = list(set(elm2 + elm1))

#         # List - None
#         elif isinstance(elm1, list) and elm2 is None:
#             for ind1, val1 in enumerate(elm1):
#                 elm1[ind1] = combine(ind1, None)

#         # if isinstance(elm1, six.string_types) and elm1 in as_shortcuts:
#         #     elm1 = as_shortcuts[elm1]

#         return elm1

#         return Environment(combine(self, service))

#     def __lt__(self, elm2):
#         return elm2 > self 




def combine(elm1, elm2, *args):

    # elm1 is merge into elm2 with elm1 as the priority
    # elm2 < elm1

    # None - None
    if elm1 is None and elm2 is None:
        return None

    # None - str
    elif elm1 is None and isinstance(elm2, six.string_types):
        elm1 = elm2

    # None - dict
    elif elm1 is None and isinstance(elm2, dict):
        elm1 = elm2
        combine(elm1, elm2)

    # None - List
    elif elm1 is None and isinstance(elm2, list):
        elm1 = elm2
        combine(elm1, elm2)

    # Dict - Dict
    elif isinstance(elm1, dict) and isinstance(elm2, dict):
        for key2, val2 in elm2.items():
            val1 = elm1.get(key2, val2)
            if val1 is not None:
                elm1[key2] = val1

    # List - List
    elif isinstance(elm1, list) and isinstance(elm2, list):
        elm1 = list(set(elm2 + elm1))

    # List - None
    elif isinstance(elm1, list) and elm2 is None:
        for ind1, val1 in enumerate(elm1):
            elm1[ind1] = combine(ind1, None)

    # if isinstance(elm1, six.string_types) and elm1 in as_shortcuts:
    #     elm1 = as_shortcuts[elm1]

    if args:
        return combine(elm1, **args)
    else:
        return elm1



import os
from docker_emperor.utils import combine, memoized_property, yaml


__all__ = ['HasEnvironment', 'Environment']


class HasEnvironment():

    preserve_environment = False
    
    @memoized_property
    def environment(self):
        if self.preserve_environment:
            self.data['environment'] = Environment(self.data.get('environment', {}))
            return self.data['environment']
        else:
            return Environment(self.data.pop('environment', {}))


class Environment(list):

    def __init__(self, value, *args, **kwargs):
        if not isinstance(value, list):
            if isinstance(value, dict):
                value = ['{}={}'.format(key,val) for key,val in value.items()]
            else:
                value = []
        self.dict = dict()
        for var in value:
            key, val = var.split('=', 1) 
            self.dict[key.strip()] = val.strip()
        list.__init__(self, value, *args, **kwargs)
    
    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.dict)

    def __gt__(self, env):
        if not isinstance(env, Environment): return self
        return env < self

    def __lt__(self, env):
        if not isinstance(env, Environment): return self
        del self[:]
        self.__init__(combine(env.dict, self.dict))
        return self

    def items(self):
        return self.dict.items()

    def __dict__(self, key):
        return self.dict

    def get(self, key, default):
        return self[key] or default

    def __getitem__(self, key):
        return self.dict.get(key)

    def __iter__(self):
        for key, val in self.items():
            yield key, val

    def __set__(self, value):
        del self[:]
        self[:] = value        

    @property
    def data(self): return self.dict


def environment_representer(dumper, environment):
    return dumper.represent_list(["{}={}".format(key, val) for key, val in environment.items()])
yaml.add_representer(Environment, environment_representer)


import six
from docker_emperor.nodes import HasEnvironment, HasServices
from docker_emperor.utils import combine, memoized_property, memoized


__all__ = ['HasContexts', 'Contexts', 'Context']


class HasContexts():
    
    @property
    def contexts(self):
        attr = '_contexts'
        if not hasattr(self, attr): 
            setattr(self, attr, Contexts(self.data.pop('contexts', {})))
        return getattr(self, attr)


class Contexts():

    def __init__(self, data):
        self.data = data
        if not isinstance(self.data, dict): self.data = {}

    @memoized
    def select(self, name='default', default_data=None): 
        data = self.data.get(name, default_data)
        if isinstance(data, six.string_types):
            name = data
            data = self.data.get(name)
        return Context(name, data)

    @property
    def default(self): 
        return self.select('default', default_data={})
    
    @property
    def all(self): 
        return { name: Context(name, data) for name, data in self.data.items() }


class Context(HasEnvironment, HasServices):

    COMMANDS = [
    ]


    def __init__(self, name, data):

        self.name = name
        self.data = data

    @classmethod
    def has_command(cls, name):
        return hasattr(cls, 'command_{}'.format(name))

    def run_command(self, name, *args):
        if Context.has_command(name):
            getattr(self, 'command_{}'.format(name))(*args)

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.name)


    # def combine_services(self):

    #     for name, service in self.data.get('services', {}).items():
    #         if service:
    #             service['environment'] = combine(service.get('environment'), self.environment, as_varlist=True)
    #             service['container_name'] = service.get('container_name', '{}.{}.{}'.format(self.project_name, self.name, name))
    #             if not 'image' in service and not 'build' in service:
    #                 service['image'] = name
    #             if 'image' in service:
    #                 if os.path.isdir(service['image']):
    #                     service['build'] = service['image']

    # @property
    # def compose_filename(self):

    #     if not hasattr(self, '_yml'):

    #         self.data.pop('shortcuts', {})

    #         self.yml = yaml.dump(self.data, Dumper=YamlDumper, default_flow_style=False, indent=4)
    #         for name, value in varlist_to_dict(self.environment).items():
    #             self.yml = self.yml.replace('${{{}}}'.format(name), value)

    #         self.compose_filename = os.path.join(self.root.root, 'docker-compose.{}.{}.yml'.format(self.project_name, self.name))
    #         self.compose_file = open(self.compose_filename, 'wb')# = tempfile.NamedTemporaryFile(mode='w+b', bufsize=-1, suffix='.yml', prefix='docker-compose-', dir=None, delete=False)
    #         self.compose_file.write(self.yml)
    #         self.compose_file.close()

    #         setattr(self, '_yml')

    #     return getattr(self, '_yml')



import six
import os
from docker_emperor.commands import Command
from docker_emperor.nodes import HasEnvironment, HasContexts, HasMachines, HasServices
from docker_emperor.utils import combine, memoized_property, yamp_dump


__all__ = ['Compose']

''' 
Compose stack
Context + Machine    
'''

class Compose(HasEnvironment, HasServices):


    def __init__(self, root, path="docker-compose"):

        self.root = root
        self.context = root.context
        self.machine = root.machine
        self.path = path
        self.name = '{}.{}'.format(self.machine.name, self.context.name)    
        self.project_name = self.data.pop(
            'project_name',  
            os.environ.get('COMPOSE_PROJECT_NAME', 
                os.path.basename(self.root.root_path)
            )
        )
        self.project_path = '{}.{}'.format(self.project_name, self.name)

    @memoized_property
    def data(self): return dict(self.root.data)

    @memoized_property
    def yml(self): 
        self.yml = yamp_dump(self.data)
        for name, value in self.environment:
            self.yml = self.yml.replace('${{{}}}'.format(name), value)
        return self.yml

    def combines(self): 
        self.environment < self.context.environment < self.machine.environment 
        self.project_name = self.environment.get('COMPOSE_PROJECT_NAME', self.project_name)
        self.services < self.context.services < self.machine.services 
        for service in self.services:
            service.environment < self.environment
            service['container_name'] = service.get('container_name', '{}.{}'.format(self.project_path, service.name))
            if not 'image' in service and not 'build' in service:
                service['image'] = service.name
            if 'image' in service:
                if os.path.isfile(os.path.join(self.root.root_path, service['image'], 'Dockerfile')):
                    service['build'] = service['image']

    @memoized_property
    def filename(self): 
        filename = os.path.join(self.root.root_path, 'docker-compose.{}.yml'.format(self.project_path))
        file = open(filename, 'wb')# = tempfile.NamedTemporaryFile(mode='w+b', bufsize=-1, suffix='.yml', prefix='docker-compose-', dir=None, delete=False)
        file.write(self.yml)
        file.close()
        return filename


    @classmethod
    def has_command(cls, name):
        return hasattr(cls, 'command_{}'.format(name))

    def run_command(self, name, *args):
        if Compose.has_command(name):
            getattr(self, 'command_{}'.format(name))(*args)


    def command_build(self, *args): 
        self.combines()
        self._run('build')

    def command_deploy(self, *args):
        target = ' docker@{}:{}'.format(self.machine.ip, os.path.join('/home/docker/', self.project_path))
        print(target)
        return Command(
            'rsync',
            '-rv --exclude ".git"',
            '--exclude ".svn"',
            '.', 
            target,
            sys=True,
            log=True
        )

    def command_run(self, *args): 
        self.combines()
        self._run('down', "--remove-orphans")
        self._run('up', *args)

    def command_rund(self, *args): 
        self.command_run('-d', *args)

    def command_down(self, *args): 
        self.combines()
        self._run('down', *args)

    def command_up(self, *args): 
        self.combines()
        self._run('up', *args)

    def command_logs(self, *args): 
        self.combines()
        self._run('logs', *args)

    def command_ps(self, *args): 
        self.combines()
        self._run('ps', *args)


    # def compose(self, *args):
    #     fn = self.context.compose_filename.replace(self.root ,'')

    #     cmd = ['docker-compose', '-f', '{} {}'.format(self.compose_filename)] + list(args)
    #     self._run(cmd)



    # @property
    # def data(self):
    #     if not hasattr(self, '_data'):

    #         data = {}

            # self.service = self.machine.merge(self.machine, self.machine)

            # data = combine(self.context.data, self.de.data)
            # self.environment = combine(self.data.pop('environment', {}), self.root.get('environment', {}), as_varlist=True)
            # self.compose_file = None
            # self.yml = ''

            # self.project_name = self.root.pop(
            #     'project_name', 
            #     self.environment.get('COMPOSE_PROJECT_NAME') if isinstance(self.environment, dict) else None
            # )
            # if not self.project_name:
            #     os.environ.get('COMPOSE_PROJECT_NAME', 'unknow')

            # self.shortcuts = combine(self.data.pop('shortcuts', {}), self.root.get('shortcuts', {}))

            # self.version = self.data.get('version', self.root.get('version', '3.3'))
            # self.data = apply_shortcuts(self.data, self.shortcuts)
            # if not 'version' in data:
            #     data['version'] = self.version
            
        #     setattr(self, '_data', data)
        # return getattr(self, '_data') 

    # def combine_services(self):

    #     for name, service in self.data.get('services', {}).items():
    #         if service:
    #             service['environment'] = combine(service.get('environment'), self.environment, as_varlist=True)
    #             service['container_name'] = service.get('container_name', '{}.{}.{}'.format(self.project_name, self.name, name))
    #             if not 'image' in service and not 'build' in service:
    #                 service['image'] = name
    #             if 'image' in service:
    #                 if os.path.isdir(service['image']):
    #                     service['build'] = service['image']

    # @property
    # def compose_filename(self):

    #     if not hasattr(self, '_yml'):

    #         self.data.pop('shortcuts', {})

    #         self.yml = yaml.dump(self.data, Dumper=YamlDumper, default_flow_style=False, indent=4)
    #         for name, value in varlist_to_dict(self.environment).items():
    #             self.yml = self.yml.replace('${{{}}}'.format(name), value)

    #         self.compose_filename = os.path.join(self.de.root, 'docker-compose.{}.{}.yml'.format(self.project_name, self.name))
    #         self.compose_file = open(self.compose_filename, 'wb')# = tempfile.NamedTemporaryFile(mode='w+b', bufsize=-1, suffix='.yml', prefix='docker-compose-', dir=None, delete=False)
    #         self.compose_file.write(self.yml)
    #         self.compose_file.close()

    #         setattr(self, '_yml')

    #     return getattr(self, '_yml')



from docker_emperor.commands import Command


def run(root, *args):

    machine_status = root.machine.status
    if machine_status == 'Stopped':
        root.machine.start()

    if machine_status == 'Running':
        root.compose.combines()
        Command(
            root.compose.path, 
            '-f', 
            root.compose.filename,
            'build',
            *args, 
            env=root.machine.docker_env,
            sys=True
        )


import os
from docker_emperor.commands import Command
import docker_emperor.logger as logger


def run(root, *args, **kwargs):

    machine_status = root.machine.status
    if machine_status == 'Stopped':
        root.machine.start()

    if machine_status == 'Running':

        # ex. docker-machine scp -r -d . virtualbox:/home/docker/project.dev.localhost/

        Command(
            root.machine.path,
            'scp',
            '-r',
            '-d',
            '.', 
            '{}:{}'.format(
                root.machine.name, 
                os.path.join('/home/docker/', root.compose.project_path)
            ),
            sys=True,
            log=True
        )

        root.run_command('build')
        root.run_command('start')



HEADER = '\033[95m'
LHEADER = '\033[1;95m'
OKBLUE = '\033[94m'
OKGREEN = '\033[92m'
GREEN = OKGREEN
WARNING = '\033[93m'
YELLOW = WARNING
LYELLOW = '\033[1;93m'
ERROR = '\033[91m'
LERROR = '\033[1;91m'
END = '\033[0m'
BOLD = '\033[1m'
UNDERLINE = '\033[4m'


def success(*value):
    value = " ".join([str(v) for v in value])
    log('{}{}{}'.format(GREEN, value, END))

def info(*value):
    value = " ".join([str(v) for v in value])
    log('{}{}{}'.format(LHEADER, value, END))

def comment(*value):
    value = " ".join([str(v) for v in value])
    log('{}{}{}'.format(HEADER, value, END))

def warning(*value):
    value = " ".join([str(v) for v in value])
    log('{}{}{}'.format(WARNING, value, END))

def error(*value):
    value = " ".join([str(v) for v in value])
    log('{}{}{}'.format(ERROR, value, END))

def log(value):
    print(value.strip())


import os
import sys
import yaml
import tempfile
import collections
import six
import argparse
import subprocess
import imp
from pprint import pprint
from docker_emperor.nodes import HasEnvironment, HasContexts, HasMachines, HasServices
from docker_emperor.compose import Compose
from docker_emperor.utils import memoized_property, combine, yamp_load
import docker_emperor.logger as logger


class DockerWorkonException(Exception):
    pass

# version, services, networks, volumes, secrets, configs, and extensions starting with "x-"


class DockerWorkon(HasEnvironment, HasServices, HasContexts, HasMachines):

    version = "1.0.0"

    def cmd_build(self, *args): 
        self.compose.combines()
        Command(
            self.compose.path, 
            '-f', 
            self.compose.filename,
            'build',
            *args, 
            env=self.machine.docker_env,
            sys=True
        )



    def __init__(self):

        self.module_root = os.path.dirname(os.path.abspath(__file__))
        self.root_path = os.getcwd()
        self.compose = None
        self.context = None
        self.machine = None 

    def entrypoint(self, action=None, *args):
        try:
            args = list(args)
            # logger.info(action, args)

            if action:

                if u'@' in action:
                    context_machine = action.split(u'@', 2)
                    self.context = self.contexts.select(context_machine[0])
                    self.machine = self.machines.select(context_machine[-1])
                    if len(args):
                        action = args[0]
                        args = args[1:]
                    else:
                        action = None
                else:
                    self.context = self.contexts.default
                    self.machine = self.machines.default

                self.compose = Compose(self)                                
                self.run_command(action, *args)

            else:
                logger.info('docker-workon version {}'.format(self.version))

        except DockerWorkonException as e:
            logger.error(e)


    def run_command(self, name, *args):

        try:
            mod = __import__('docker_emperor.commands.{}'.format(name), globals(), locals(), ['run'], 0)
            mod.run(self, *args)
        except ImportError as e:
            logger.error('Unknown command {}'.format(name))
        # if hasattr(self, cmd_method):
        #     getattr(self, cmd_method)(*args, **kwargs)

    @memoized_property
    def data(self):
        for file in ['docker-emperor.yml', 'docker-compose.yml']:
            filename = os.path.join(self.root_path, file)
            if os.path.isfile(filename):
                data = yamp_load(open(filename, 'rb').read())
                if not isinstance(data, dict):
                    raise DockerWorkonException('{} is not yml as dict'.format(os.path.basename(file)))
                return data
        raise DockerWorkonException('{} not found in {}'.format(os.path.basename(file), self.root_path))


dw = DockerWorkon()

def entrypoint():
    argsparser = argparse.ArgumentParser(description='Docker web apps composer')
    argsparser.add_argument('args', nargs=argparse.REMAINDER, action="store")    
    dw.entrypoint(*argsparser.parse_args().args)

if __name__ == "__main__": entrypoint()










# # @property
# # def machine(self):
# #     attr = '_machine'
# #     if not hasattr(self, attr): setattr(self, attr, self.machines.select(self.machine_name))
# #     return getattr(self, attr)


# def _run(self, *args):
#     return Command(self, self.compose_path, *args, env=self.machine.docker_env)


# def cmd_start(self, *args):
#     self.cmd_prune()
#     self.compose("down --remove-orphans")
#     self.compose("up", *args)

# def cmd_startd(self, *args):
#     self.cmd_start('-d', *args)

# def cmd_logs(self, *args):
#     self.compose('logs', '-f', *args)


# def cmd_run(self, *args):
#     self.compose('run', *args)

# def cmd_exec(self, *args):
#     self.compose('exec', *args)

# def cmd_sethost(self, host):
#     bin_path = os.path.join(self.module_root, 'bin')
#     self.cmd("docker run -t -i -v {bin_path}/sethost:/bin/sethost -v /etc/hosts:/etc/hosthosts -v /etc/hosts.bak:/etc/hosthosts.bak busybox sh /bin/sethost 0.0.0.0 {}".format(host))

# def cmd_prune(self):
#     self.cmd("docker network prune -f")
#     self.cmd("docker system prune -f")
#     self.cmd("docker volume prune -f")


# def cmd(self, *args):
#     cmd = " ".join(args).strip()
#     self.log_info('> {}'.format(cmd))
#     try:
#         output = subprocess.check_output(cmd.split())
#         print(output.strip())
#     except Exception as e:
#         raise DockerWorkonException()#"error: " + str(e))#, sys.exc_info())


# def log_info(self, message):
#     print('{}{}{}{}'.format(C.YELLOW, C.LYELLOW, message, C.ENDC))

# def log_success(self, message):
#     print('{}{}{}{}'.format(C.GREEN, C.LGREEN, message, C.ENDC))

# def log_error(self, message):
#     print('{}{}{}{}'.format(C.ERROR, C.LERROR, message, C.ENDC))





import os
from docker_emperor.commands import Command
import docker_emperor.logger as logger


def run(root, *args, **kwargs):

    Command(
        root.compose.path,
        'down',
        '--remove-orphans',
        sys=True,
    )